//#define DEBUG_SPD 1

#include "spd.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

SpeechdClient::SpeechdClient(const char *clientName, const char *host, int port, bool interactive)
{
    
    _host=host?strdup(host):nullptr;
    _port=port?port:DEFAULT_PORT;
    is_connected=0;
    _interactive=interactive;
    
    if (!clientName) _clientName=strdup(ESPSTRING ":" ESPSTRING ":main");
    else {
        const char *c =strchr(clientName,':');
        if (!c) sprintf(linebuf,ESPSTRING ":" ESPSTRING ":%s", clientName);
        else {
            c=strchr(c+1,':');
            if (c) strcpy(linebuf, clientName);
            else sprintf(linebuf,ESPSTRING ":%s", clientName);
        }
        _clientName = strdup(linebuf);
    }
    _rate=0;
    _module[0]=0;
    _language[0]=0;
    _voice[0]=0;
}

SpeechdClient::~SpeechdClient(void)
{
    if (is_connected) client.stop();
    free(_host);
    free(_clientName);
}

int SpeechdClient::readline(char *buf, int maxlen)
{
    uint32_t started = millis();
    int i=0,n;
    if (!buf) {
        buf = linebuf;
        maxlen=256;
    }
    for (;;) {
        while (!client.available()) {
            if (millis() - started > 3000UL) return -1;
            delay(2);
        }
        while (client.available()) {
            n=client.read();
            if (n == '\r') continue;
            if (n == '\n') {
                buf[i] = 0;
#ifdef DEBUG_SPD
                printf("<- %s\n", buf);
#endif
                return 0;
            }
            if (i < maxlen-1) buf[i++]=n;
        }
    }
}

int SpeechdClient::writeln(const char *str, size_t len)
{
    if (!len) len=strlen(str);
    if (!is_connected) return -1;
#ifdef DEBUG_SPD
    printf("-> %-*.*s\n",len, len, str);
#endif
    while (len-- > 0) {
        if (!client.write(*str++)) break;
    }
    if (client.print("\r\n") != 2) return -1;
    return 0;
}

int SpeechdClient::stop(void)
{
    if (!is_connected) return 0;
    if (writeln("CANCEL self")) return -1;
    return response();
}

int SpeechdClient::_setParam(const char *param, const char *value)
{
    sprintf(linebuf,"SET self %s %s", param, value);
    if (writeln(linebuf)) return -1;
    return response();
}

int SpeechdClient::_setIParam(const char *param, int value)
{
    sprintf(linebuf,"SET self %s %d", param, value);
    if (writeln(linebuf)) return -1;
    return response();
}

void SpeechdClient::disconnect(void)
{
    if (is_connected) client.stop();
    is_connected=0;
}

int SpeechdClient::connect(void)
{
    int n,clid;
    if (is_connected) return 0;
    if (!_host) return -1;
    if (!client.connect(_host, _port ? _port : DEFAULT_PORT)) {
        return -1;
    }
    is_connected=1;
    _messageID=0;
    sprintf(linebuf,"SET self CLIENT_NAME %s", _clientName);
    if (writeln(linebuf)) return -1;
    if ((n=response())) return n;
    if (writeln("HISTORY GET CLIENT_ID")) return -1;
    if ((n=response(&clid))) return n;
    clientID=clid;
#ifdef DEBUG_SPD
    printf("Client ID: %d\n", clid);
#endif
    if (_interactive) {
        if ((n=_setParam("NOTIFICATION BEGIN", "on"))) return n;
        if ((n=_setParam("NOTIFICATION END", "on"))) return n;
        if ((n=_setParam("NOTIFICATION CANCEL", "on"))) return n;
    }
    if (_module[0]) {
        if ((n=_setParam("OUTPUT_MODULE",_module))) return n;
    }
    if (_language[0]) {
        if ((n=_setParam("LANGUAGE", _language))) return n;
    }
    if (_voice[0]) {
        if ((n=_setParam("SYNTHESIS_VOICE", _voice))) return n;
    }
    if (_rate) {
        char buf[10];
        sprintf(buf,"%d", _rate);
        if ((n=_setParam("RATE", buf))) return n;
    }
    return 0;
}

int SpeechdClient::setModule(const char *module)
{
    strcpy(_module, module);
    return is_connected?_setParam("OUTPUT_MODULE",_module):0;
}

int SpeechdClient::setVoice(const char *voice)
{
    strcpy(_voice, voice);
    return is_connected?_setParam("SYNTHESIS_VOICE",_voice):0;
}

int SpeechdClient::setRate(int rate)
{
    _rate = rate;
    if (is_connected && _rate) {
        char buf[10];
        sprintf(buf,"%d", _rate);
        return _setParam("RATE", buf);
    }
    return 0;
}

int SpeechdClient::setLanguage(const char *language)
{
    strcpy(_language,language);
    return is_connected?_setParam("LANGUAGE", language):0;
}

int SpeechdClient::say(const char *str, int priority)
{
    int n=_say(str, priority);
    if (!n) return 0;
    if (is_connected) {
        client.stop();
        is_connected=0;
    }
#ifdef DEBUG_SPD
    if (n > 0) printf("ERROR %s\n", linebuf);else if (n < 0) printf("ERROR %d: connection error\n",n);
#endif
    return n;
}

int SpeechdClient::_say(const char *str, int priority)
{
    int n,msgid=0;
    const char *c;
    static const char *prios[]={"progress","notification","text","message","important"};
    if (connect()) return -1;
    if (writeln("CANCEL self")) return -1;
    if ((n=response())) return n;
    if (priority > 0 && priority < PRIO_MAX) {
        sprintf(linebuf,"SET self PRIORITY %s", prios[priority-1]);
        if (writeln(linebuf)) return -1;
        if ((n=response())) return n;
    }
    if (writeln("SPEAK")) return -1;
    if ((n=response())) return n;
    while (*str) {
        if (*str == '.') {
            str++;
            continue;
        }
        c=strpbrk(str,"\r\n");
        if (!c) c = str+strlen(str);
        if (writeln(str, c-str)) return -1;
        str=c;
    }
    if (writeln(".")) return -1;
    n=response(&msgid);
    if (n) return n;
    _messageID = msgid;
    _messageStatus = 1; // queued
#ifdef DEBUG_SPD
    printf("Message id %d\n", msgid);
#endif
    return 0;
}

int SpeechdClient::status(void)
{
    if (!is_connected) return 0;
    if (!_interactive) return -1;
    if (!_messageID) return 0;
    return _messageStatus;
}
int SpeechdClient::_readNotify(void)
{
    int params[2];
    int numpar=0;
    int n;
    int lcode;
    char *c;
    for (;;) {
        lcode = strtol(linebuf,&c,10);
        if (lcode <700 || lcode >=800) return lcode;
        if (*c++ != '-') break;
        if (numpar < 2) params[numpar++] = strtol(c,NULL,10);
        if ((n=readline())) return n;
    }
    if (numpar < 2 || params[1] != clientID || params[0] != _messageID) return 0;
    if (lcode == 702 || lcode == 703) {
        _messageID=0;
    }
    else if (lcode == 701) {
        _messageStatus = 2;
    }
#ifdef DEBUG_SPD
    printf("NOTIFIED message %d: %s\n", params[0], linebuf);
#endif
    return 0;
}

int SpeechdClient::loop(void)
{
    int n;
    if (!is_connected) return 0;
    if (!_interactive) return 0;
    if (!client.connected()) {
        client.stop();
        is_connected=0;
        return -1;
    }
    for (;;) {
        if (!client.available()) return 0;
        if ((n=readline())) break;
        if ((n=_readNotify())) break;
    }
    client.stop();
    is_connected=0;
    return n;
}

int SpeechdClient::response(int *msg)
{
    char *c;
    int n;
    if (!is_connected) return -1;
    for (;;) {
        if (readline()) return -1;
        n=strtol(linebuf, &c, 10);
        if (n >=700 && n < 800) {
            if ((n=_readNotify())) return n;
            continue;
        }
        if (n < 200 || n >= 300) return n;
        if (*c++ == '-') {
            if (msg) *msg=strtol(c,&c,10);
        }
        else break;
        //if (*c != '-') break;
    }
    return 0;
}

int SpeechdClient::setHost(uint32_t host, int port)
{
    char buf[20];
    sprintf(buf,"%d.%d.%d.%d",(host>>0) & 255, (host>>8) & 255, (host>>16) & 255, (host>>24) & 255);
    return setHost(buf, port);
}

int SpeechdClient::setHost(const char *host, int port)
{
    if (!host || !*host || *host==':') return -1;
    char *c=strchr(host,':');
    char *h = nullptr;
    int p = port;
    if (c) {
        p=strtol(c+1,&c,10);
        if (*c) return -1;
        int len=c-host;
        h=(char *)malloc(len+1);
        memcpy((void *)h, (void *)host, len);
        h[len]=0;
    }
    else {
        h=strdup(host);
    }
    if (is_connected) {
        is_connected=0;
        client.stop();
    }
    if (_host) free(_host);
    _host=h;
    _port=p;
    return 0;
}
